/*  Copyright 2008 Peter Mora
 *
 *  This file is part of Vakarcs.
 *
 *  Vakarcs is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Vakarcs is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Vakarcs.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
//#include <pthread.h>
#ifdef __WIN32__
 #include "socket/tcpsocket.hh"
#else
 #include "tcpsocket.hh"
#endif
#include "server.h"
#include <stdlib.h>

int main(int argc, char **argv)
{
	int			port;
	if (argc > 3 )
	{
		std::cerr << "Usage : " << argv[0] << std::endl;
		std::cerr << "or : " << argv[0] << " number_of_players" << std::endl;
		std::cerr << "or : " << argv[0] << " number_of_players port" << std::endl;
		return 1;
	}
	if (argc == 3)
		port = strtol(argv[2], 0, 10);
	else
		port = 5555;
	if (port <= 0)
	{
		std::cerr << "Invalid port" << std::endl;
		return 1;
	}
	int number_of_players = 2;
	if (argc >= 2)
		number_of_players = strtol(argv[1], 0, 10);
	vector<unsigned int> player_ids(number_of_players);
	vector<int> player_ready(number_of_players);
	for (int i = 0; i < number_of_players; i++)
		player_ids[i] = i;
	int found_players = 0;

	unsigned int game_id = rand();
	char game_id_str[50];
	sprintf(game_id_str,"%d",game_id);

	char tmp[50];

	Server server;

	try
	{
		Network::TcpSocket	tcpserver, *tcpclient;

		try
		{
			tcpserver.connect(port);
			tcpserver.add_delim("\n");
			tcpserver.add_delim("\r\n");
		}
		catch (Network::Exception &e)
		{
			std::cerr << e;
			return 1;
		}
		std::cout << "Opening port " << port << std::endl;
		std::cout << "Waiting..." << std::endl;


		string first_message = server.init(number_of_players);

		while (found_players != number_of_players)
		{
			//	  if (argc > 2)
			//	    tcpserver.init_tls(Network::LIBSOCKET_TLS, 1024, argv[2], argv[3]);

//			unsigned int server_code = rand();
			tcpclient = tcpserver.accept();
			tcpclient->add_delim("\n");
			tcpclient->add_delim("\r\n");
			std::string str;
			(*tcpclient) >> str;
			if (str == "WANNA PLAY")
			{
				std::cout << "Incomming connection from " << tcpserver.get_ip(tcpclient) << std::endl;
				(*tcpclient) >> str;
				if (str != VERSION)
				{
					std::cout << "The client uses version " << str << ", but we use version " << VERSION <<". Client dropped." << std::endl;
				}
				else
				{
					std::cout << found_players+1 << "/" << number_of_players << " player(s) has connected!" << std::endl;
					(*tcpclient) << "Let's start\n";
					sprintf(tmp,"%d",player_ids[found_players++]);
					(*tcpclient) << tmp << "\n";
					sprintf(tmp,"%d",number_of_players);
					(*tcpclient) << tmp << "\n";
					(*tcpclient) << game_id_str << "\n";
					(*tcpclient) << itostr(server.maxx()) << "\n";
					(*tcpclient) << itostr(server.maxy()) << "\n";
					(*tcpclient) << first_message << "\n";
				}
			}
			if (str == "Give me data")
			{
				(*tcpclient) >> str;
				(*tcpclient) >> str;
				sprintf(tmp,"%d",number_of_players-found_players);
				if (number_of_players-found_players == 1)
					(*tcpclient) << "We need " << tmp << " more player!\n";
				else
					(*tcpclient) << "We need " << tmp << " more players!\n";
			}

			tcpclient->close();
		}
		
		//We have all the players!

		for (int i= 0; i < number_of_players; i++)
			player_ready[i] = 0;

		std::cout << "Waiting..." << std::endl;
		bool	is_everybody_has_one = false;
		while (server.is_gameover() == false || is_everybody_has_one == false)
		{

//			unsigned int server_code = rand();
			tcpclient = tcpserver.accept();
			tcpclient->add_delim("\n");
			tcpclient->add_delim("\r\n");
			std::string str;
			(*tcpclient) >> str;
			if (str == "WANNA PLAY")
			{
				(*tcpclient) << "Server is full.\n";
			}
			if (str == "Give me data")
			{
				(*tcpclient) >> str;
				int player_id;
				player_id = strtol(str.c_str(),0,10);
				(*tcpclient) >> str;
				if ( strtol(str.c_str(),0,10) != game_id)
				{
					usleep(300000);
					(*tcpclient) << "Unknown game id\n";
				}
				else
				{
					if (player_ready[player_id] == 0)
					{
						(*tcpclient) << "Transmitting data.\n";
						(*tcpclient) << server.send(player_id) << "\n";
						player_ready[player_id] = 1;
						if (server.active_players()[player_id] == false)
							player_ready[player_id] = -1;
					}
					else
					{
						if (player_ready[player_id] == 1)
							(*tcpclient) << "What are your steps?\n";
						else
						{
							(*tcpclient) << "Waiting for the following player(s): ";
							for (int i = 0; i < number_of_players; i++)
							{
								if (player_ready[i] == 1)
									(*tcpclient) << itostr(i+1) << " ";
							}
							(*tcpclient) << "\n";
						}
					}
				}
			}
			if (str == "Transmitting data.")
			{
				(*tcpclient) >> str;
				int player_id;
				player_id = strtol(str.c_str(),0,10);
				(*tcpclient) >> str;
				if ( strtol(str.c_str(),0,10) != game_id)
				{
					usleep(300000);
					(*tcpclient) << "Unknown game id\n";
				}
				else
				{
					if (player_ready[player_id] == 1)
					{
						(*tcpclient) << "Receiving.\n";
						(*tcpclient) >> str;
						if (str == "exit")
						{
							player_ready[player_id] = -1;
							server.leaver(player_id);
							bool end = true;
							for (int i = 0; i < number_of_players; i++)
								if (player_ready[i] != -1)
								{
									end = false;
									break;
								}
							if (end)
								return 0;

						}
						else
						{
							server.receive(player_id,str);
							player_ready[player_id] = 2;
						}
					}
					else
					{
						if (player_ready[player_id] == 0)
							(*tcpclient) << "First, download the new data.\n";
						else
							(*tcpclient) << "You have already sent your moves.\n";
					}

				}
			}

			bool finish_round = true;
			is_everybody_has_one = true;
			for (int i = 0; i < number_of_players; i++)
			{
				if (player_ready[i] != 2 && player_ready[i] != -1)
					finish_round=false;
				if (player_ready[i] < 1 && player_ready[i] != -1)
					is_everybody_has_one = false;
			}
			if (finish_round)
			{
				for (int i = 0; i < number_of_players; i++)
					if (player_ready[i] != -1)
						player_ready[i] = 0;
				server.make_rounds();
				is_everybody_has_one = false;
			}
			tcpclient->close();
		}



	}
	catch (Network::Exception &e)
	{
		std::cerr << e;
		return 1;
	}
	return 0;
}
