/*  Copyright 2008 Peter Mora
 *
 *  This file is part of Vakarcs.
 *
 *  Vakarcs is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Vakarcs is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Vakarcs.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.h"
#include <iostream>
#include <stdlib.h>

Server::Server()
{
	srand(time(NULL));
	m_round_time = 10;
	m_first_info_player_not_has = NULL;
	m_last_id = 0;
}

Server::~Server()
{
	if (m_first_info_player_not_has)
		delete[] m_first_info_player_not_has;
	delete[] m_punishment_from_timeback;
	delete[] m_active_players;
	for (int i = 0; i < m_number_of_players; i++)
		for (int j = 0; j < m_player_info[i].size(); j++)
			if (m_player_info[i][j])
			{
				delete[] m_player_info[i][j]->map_for_player;
				delete[] m_player_info[i][j]->score;
				for (int q = 0; q < m_player_info[i][j]->objects.size(); q++)
					delete m_player_info[i][j]->objects[q];
				delete m_player_info[i][j];
			}
	delete[] m_player_info;
	for (int i = 0; i < m_next_player_info.size(); i++)
		delete m_next_player_info[i];
	delete[] m_bases;
	for (int i = 0; i < m_round.size(); i++)
	{
		delete[] m_round[i]->map;
		delete[] m_round[i]->visible_maps_for_players;
		delete[] m_round[i]->score;
		for (int j = 0; j < m_round[i]->objects.size(); j++)
			delete m_round[i]->objects[j];
		delete m_round[i];
	}
	delete[] m_mirror;
	for (int i = 0; i < m_abilities.size(); i++)
		delete m_abilities[i];
	delete[] m_names;
}


string Server::init(int n)
{
	m_number_of_players = n;
	m_punishment_from_timeback = new int[m_number_of_players];
	for (int i = 0; i < m_number_of_players; i++)
		m_punishment_from_timeback[i]=0;
	for (int i = 0; i < m_number_of_players; i++)
		m_next_player_info.push_back(new PlayerInfo);
	m_player_info = new vector<PlayerInfo*>[m_number_of_players];
	m_first_info_player_not_has = new int[m_number_of_players];
	m_active_players = new bool[m_number_of_players];
	m_game_over = false;
	m_bases = new int[2*m_number_of_players];
	for (int i = 0; i < m_number_of_players; i++)
	{
		m_first_info_player_not_has[i] = 0;
		m_active_players[i] = true;
	}
	m_incoming_messages.resize(n);
	m_outgoing_messages.resize(n);
	std::string str;
	char s[1000];
	sprintf(s,"%d",n);
	str = s;
	str = "startscript"+str;
	str = str+".txt";
	FILE* f;
	if ( (f=fopen(str.c_str(),"r"))==NULL)
	{
		std::cout << "Error opening file: " << str << std::endl;
		exit(1);
	}
	Round* r = new Round;
//	m_maxx = 80;
//	m_maxy = 22;


	m_round.push_back(r);
	m_copy_round_from  = 0;
	fscanf(f,"%d, %d\n",&m_maxx, &m_maxy);
	m_mirror = new int[m_number_of_players];
	for (int i = 0; i < m_number_of_players; i++)
	{
		m_mirror[i]= rand()%(m_maxx==m_maxy?8:4);
	}
	r->map = new int[m_maxy*m_maxx];
	for( int i = 0; i < m_maxx*m_maxy; i++)
		r->map[i] = 0;
	r->visible_maps_for_players = new int[m_maxy*m_maxx*m_number_of_players];
	for( int i = 0; i < m_maxx*m_maxy*m_number_of_players; i++)
		r->visible_maps_for_players[i] = MAPS_VISIBLE;
	for (int y = 0; y < m_maxy; y++)
	{
		fscanf(f,"%s\n",s);	// TODO: use a safe read mode
		if (strlen(s) < m_maxx)
		{
			cout << "Error with startscript file.\n";
			exit(1);
		}
		for (int x=0; x < m_maxx; x++)
		{
			switch (s[x])
			{
				case '`':
					r->map[y*m_maxx+x] = 0;
					break;
				case '*':
					r->map[y*m_maxx+x] = MAP_WALL;
					break;
			}
			if (s[x] >= 'A' && s[x] <= 'Z')
			{
				r->map[y*m_maxx+x] = MAP_BASE + (int)(s[x] - 'A');
				if ((s[x]-'A') < m_number_of_players)
				{
					m_bases[2*(s[x]-'A')] = x;
					m_bases[2*(s[x]-'A')+1] = y;
				}
			}
		}
	}

	string ret; // we will return this string

	fscanf(f,"start points: %d\n",&m_to_win);
	int start_vak_number;
	fscanf(f,"start vakarcs number: %d\n",&start_vak_number);
	fscanf(f,"vakarcs at every nth round: %d\n",&m_vak_number);
	m_vak_number_down = m_vak_number;
	fscanf(f,"abilities: %s\n",s);
	m_abilities_types = s;
	ret = m_abilities_types + "^";
	r->score = new int[m_number_of_players];
	for (int i =0; i < m_number_of_players; i++)
		r->score[i] = m_to_win;
/*	while(!feof(f))
	{
		int i;
		fscanf(f,"%d ",&i);
		fgets(s,1000,f);
		if (feof(f))
			break;
		Startscript* ss = new Startscript;
		ss->object = s;
		ss->round = i;
		m_startscript.push_back(ss);
	}
*/
//	bool first = true;
	while(!feof(f))
	{
		int a,b,c,d;
		fscanf(f,"%d ",&a);
		fscanf(f,"%d ",&b);
		fscanf(f,"%d ",&c);
		fscanf(f,"%d ",&d);
		fgets(s,1000,f);
		if (feof(f))
			break;
		Ability* ab = new Ability;
		for (int i = 0; i < 10; i++)
			ab->number[i] = -1;
		ab->red = a;
		ab->green = b;
		ab->blue = c;
		ab->white = d;
		vector<std::string>* t = tokenize(s,' ',true);
//		if (!first)
//			ret += "^";
//		first = false;
		ret += itostr(a) + " " + itostr(b) + " " + itostr(c) + " "+ itostr(d) + " ";
		int j = 0;
		while (s[j] && s[j] != 10 && s[j] != 13)
			ret+= s[j++];
		ret += " ";
		for(int i = 0; i < t->size(); i+=3)
		{
			int w = (*t)[i][0]-48;
			if (w >= 0 && w < 10)
			{
				ab->id.push_back(w);
				ab->number[w] = ab->ability.size();
			}
			else
				ab->id.push_back(0);
			ab->ability.push_back((*t)[i].substr(1,(*t)[i].size()-1));
			ab->parameter.push_back(atol((*t)[i+1].c_str()));
			ab->cooldown.push_back(atol((*t)[i+2].c_str()));
		}
		m_abilities.push_back(ab);
		delete t;
		ret += "^";
	}

	fclose(f);
	if ( (f=fopen("names.txt","r"))==NULL)
	{
		cout << "Error, file names.txt is missing.\n";
		exit(1);
	}
	m_names = new vector<string>[m_number_of_players];
	while(!feof(f))
	{
		fgets(s,1000,f);
		if (feof(f))
			break;
		for (int i = 0; s[i]; i++)
			if (s[i] == 13 || s[i] == 10)
				s[i] = 0;
		vector<string>* t = tokenize(s,'\t',true);

		for (int i = 0; i < t->size(); i++)
			if (i < m_number_of_players)
			{
				m_names[i].push_back((*t)[i]);
			}

		delete t;
	}
	for (int j = 0; j < m_number_of_players; j++)
	{
		if (m_names[j].size() == 0)
		{
			m_names[j].push_back("Joe");
			m_names[j].push_back("Jack");
		}
		random_shuffle(m_names[j].begin(),m_names[j].end());
	}
/*	for (int j = 0; j < m_startscript.size(); j++)
		if (m_startscript[j]->round == 0)
		{
			string t = m_startscript[j]->object;
			for (int q = 0; q < COOLDOWN; q++)
				t += " 0";
			r->objects.push_back(new Object(t));
		}
*/
	for (int w = 0; w < start_vak_number; w++)
		for (int i = 0; i < m_number_of_players; i++)
		{
			string t = itostr(m_last_id++);
			t += " ";
			t += m_names[i].front();
			if (m_number_of_players > 2)
				t += "(" + itostr(i+1) + ")";
			rotate(m_names[i].begin(),m_names[i].begin()+1,m_names[i].end());
			int q = rand() % 4;
			for (int w = 0; w < 4; w++)
			{
				if (q == w)
					t += " 1";
				else
					t +=" 0";
			}
			t += " ";
			t += itostr(i) + " ";
			t += itostr(m_bases[2*i]) + " " + itostr(m_bases[2*i+1]) + " ";
			t += "1 0 0 ";
			for (int q = 0; q < COOLDOWN; q++)
				t += "0 ";
			r->objects.push_back(new Object(t));
		}

	for (int i = 0; i < m_number_of_players; i++)
	{
		PlayerInfo* player = new PlayerInfo;
		player->round = 0;
		player->strings.push_back(string("Welcome!"));
		player->map_for_player = new int[m_maxx*m_maxy];
		player->score = new int[m_number_of_players];
		for (int j = 0; j < m_number_of_players; j++)
			player->score[j] = r->score[j];
		for (int x = 0; x < m_maxx; x++)
			for (int y =0; y < m_maxy; y++)
				player->map_for_player[y*m_maxx+x] = r->map[y*m_maxx+x];
		for (int j = 0; j < r->objects.size(); j++)
			if (r->objects[j]->owner() == i)
			{
				player->objects.push_back(new Object(r->objects[j]));
			}
		m_player_info[i].push_back(player);
	}
	generate_messages();
	return ret;
}
std::string Server::send(int playerid)
{
	return m_outgoing_messages[playerid];
}
void	Server::receive(int playerid, std::string str)
{
	m_incoming_messages[playerid] = str;
}

void	Server::make_rounds()
{
	for (int i = 0 ; i < m_number_of_players; i++)
		m_outgoing_messages[i]=string("");
	if (m_game_over)
		return;

	vector<string>** mes = new vector<string>*[m_number_of_players];
	for (int i = 0 ; i < m_number_of_players; i++)
	{
		mes[i] = tokenize(m_incoming_messages[i],'$');
	}

	for (int i = 0; i < m_round_time && !m_game_over; i++)  //ROUNDS
	{
		Round* last = m_round[m_copy_round_from];//.back();

		Round* r = new Round;
		r->score = new int[m_number_of_players];
		for (int j = 0; j < m_number_of_players; j++)
			r->score[j] = last->score[j];
		r->map = new int[m_maxy*m_maxx];
		for( int j = 0; j < m_maxx*m_maxy; j++)
		{
			r->map[j] = last->map[j];
			if (r->map[j] == MAP_DESTROYED)
				r->map[j] = 0;
		}
		r->visible_maps_for_players = new int[m_maxy*m_maxx*m_number_of_players];
		for( int m = 0; m < m_maxx*m_maxy*m_number_of_players; m++)
			r->visible_maps_for_players[m] = MAPS_BLIND;

		vector<Object*>* obj = &(last->objects);
		for (int j = 0; j < obj->size(); j++)
			if (m_active_players[last->objects[j]->owner()])
				r->objects.push_back(new Object(last->objects[j]));

		for (int j = 0; j < m_startscript.size(); j++)
			if (m_startscript[j]->round == m_round.size())
			{
				r->objects.push_back(new Object(m_startscript[j]->object));
			}
//		if (i == m_round_time -1)
//			for (int w = 0; w < m_vak_number; w++)
		for (int j = 0; j < m_number_of_players; j++)
			if (mes[j]->size() > i && m_active_players[j])
			{
				vector<string>* vak = tokenize((*mes[j])[i],'#');
				for (int k = 0; k < vak->size(); k++)
				{
					set_up_step(r, (*vak)[k], j);
				}
				delete vak;
			}
		m_round.push_back(r);
		m_copy_round_from  = m_round.size() -1;
		make_step();
	}
	for (int i = 0 ; i < m_number_of_players; i++)
		delete mes[i];
	delete[] mes;
	if (--m_vak_number_down == 0)
	{
		for (int j = 0; j < m_number_of_players; j++)
			if (m_active_players[j])
			{
				string t = itostr(m_last_id++);
				t += " ";
				t += m_names[j].front();
				if (m_number_of_players > 2)
					t += "(" + itostr(j+1) + ")";
				rotate(m_names[j].begin(),m_names[j].begin()+1,m_names[j].end());
				int q = rand() % 4;
				for (int w = 0; w < 4; w++)
				{
					if (q == w)
						t += " 1";
					else
						t +=" 0";
				}
				t += " ";
				t += itostr(j) + " ";
				t += itostr(m_bases[2*j]) + " " + itostr(m_bases[2*j+1]) + " ";
				t += "1 0 0 ";
				for (int q = 0; q < COOLDOWN; q++)
					t += "0 ";
				m_round.back()->objects.push_back(new Object(t));
				m_player_info[j][m_player_info[j].size()-1]->objects.push_back(new Object(t));
			}
		m_vak_number_down = m_vak_number;
	}
	generate_messages();
};

void	Server::generate_messages()
{
	// Generate outgoing messages
	for (int i =0; i < m_number_of_players; i++)
	{
		for(;m_first_info_player_not_has[i] < m_player_info[i].size(); m_first_info_player_not_has[i]++)
		{
			PlayerInfo* p = m_player_info[i][m_first_info_player_not_has[i]];
			m_outgoing_messages[i] += itostr(p->round) + string("#");
			if (m_active_players[i] && !m_game_over) //m_game_over)
				m_outgoing_messages[i] += "0" + string("#");
			else
				m_outgoing_messages[i] += "1" + string("#");
			//cout << i << " " << p->strings.size() << "\n";
			//cout.flush();
			
			for (int j = 0; j < p->strings.size(); j++)
			{
				m_outgoing_messages[i] += p->strings[j];
				m_outgoing_messages[i] += string("^");
			}
			m_outgoing_messages[i] += string("#");
			for (int y= 0 ; y < m_maxy; y++)
				for (int x = 0; x < m_maxx; x++)
					m_outgoing_messages[i] += itostr(p->map_for_player[mirrory(x,y,i)*m_maxx+mirrorx(x,y,i)]>=MAP_BASE?MAP_BASE:p->map_for_player[mirrory(x,y,i)*m_maxx+mirrorx(x,y,i)])+string("^");
			m_outgoing_messages[i] += string("#");
			for (int j = 0; j < p->objects.size(); j++)
			{
				m_outgoing_messages[i] += p->objects[j]->generate_string(m_mirror[i],m_maxx,m_maxy);
//				if (j < p->objects.size()-1)
					m_outgoing_messages[i] += string("^");
			}
			m_outgoing_messages[i] += string("#");
			for (int j = 0; j < m_number_of_players; j++)
			{
				m_outgoing_messages[i] += itostr(p->score[j]-m_punishment_from_timeback[j]);
//				if (j < m_number_of_players-1)
					m_outgoing_messages[i] += "^";
			}
			m_outgoing_messages[i] += string("#");
//			if (m_first_info_player_not_has[i] < m_player_info[i].size() -1)
				m_outgoing_messages[i] += string("$");
		}
//		cout << m_outgoing_messages[i] << "\n";
	}
}

void	Server::make_step()
{
	vector<Object*>* obj = &(m_round.back()->objects);
	for(int k = 0; k < obj->size(); k++)
		(*obj)[k]->action(0,this,NULL,NULL,NULL);
	for (int j = 1; j < 100; j++)
	{
		vector<Object*> actives;
		for(int k = 0; k < obj->size(); k++)
		{
			if ((*obj)[k]->active()>0)
				actives.push_back((*obj)[k]);
		}

		bool* generate = new bool[m_number_of_players];
	       	for (int i = 0; i < m_number_of_players; i++)
			generate[i] = false;
		bool* redraw = new bool[m_number_of_players];
		for(int k = 0; k < actives.size(); k++)
		{
			for (int i = 0; i < m_number_of_players; i++)
				redraw[i] = false;
			actives[k]->action(j,this,redraw,(m_next_player_info[actives[k]->owner()]),&m_abilities);
			for (int i = 0; i < m_number_of_players; i++)
				if (redraw[i] || j == 40)
					generate[i] = true;
		}
		for (int i = 0; i < m_number_of_players; i++)
			if (generate[i])
				generate_maps(i);
		delete[] generate;
		delete[] redraw;
	}
	//push all the messages into last round
	
	check_game_over();

	for (int i = 0 ; i < m_number_of_players; i++)
	{
		for (int j = 0; j < m_next_player_info[i]->strings.size(); j++)
			m_player_info[i][m_player_info[i].size()-1]->strings.push_back(m_next_player_info[i]->strings[j]);
		m_next_player_info[i]->strings.clear();
	}
}

void	Server::generate_maps(int playerid)
{
	Round* last = m_round.back();
	PlayerInfo* player = m_next_player_info[playerid];
        m_next_player_info[playerid] = new PlayerInfo;
//	for (int i =0; i < m_number_of_players; i++)
	{		
		m_player_info[playerid].push_back(player);
		player->round = m_round.size()-1;
		int *to_load = new int[m_maxx*m_maxy];
		int* lastmap = m_player_info[playerid][m_player_info[playerid].size()-2]->map_for_player;
		player->map_for_player = to_load;
		player->score = new int[m_number_of_players];
		for (int j = 0; j < m_number_of_players; j++)
			player->score[j] = last->score[j];

		for (int x = 0; x < m_maxx; x++)
			for (int y = 0; y < m_maxy; y++)
			{
				to_load[y*m_maxx+x] = lastmap[y*m_maxx+x];
				if (to_load[y*m_maxx+x]==MAP_EMPTY || to_load[y*m_maxx+x]==MAP_DESTROYED )
					to_load[y*m_maxx+x] = 0;
				if (last->visible_maps_for_players[m_maxx*m_maxy*playerid + y*m_maxx+x] == MAPS_VISIBLE || last->map[y*m_maxx+x]==MAP_DESTROYED)
				{
					to_load[y*m_maxx+x] = last->map[y*m_maxx+x];
					if (to_load[y*m_maxx+x] == 0)
						to_load[y*m_maxx+x] = MAP_EMPTY;
				}
			}
		for (int j = 0; j < last->objects.size(); j++)
			if ((last->objects[j]->owner() == playerid || last->visible_maps_for_players[m_maxx*m_maxy*playerid + m_maxx*last->objects[j]->y()+last->objects[j]->x()] == MAPS_VISIBLE || last->objects[j]->visible() > 0) && last->objects[j]->active()>0)
				player->objects.push_back(new Object(last->objects[j]));

	}
}

void	Server::set_up_step(Round* r, string step, int owner)
{
	vector<string>*	s = tokenize(step,' ');
	if (s->size()!= 4)
		return;
	int l = atol((*s)[0].c_str());
	string type = (*s)[1];
	int x = atol((*s)[2].c_str());
	int y = atol((*s)[3].c_str());

	for (int i = 0; i < r->objects.size(); i++)
		if (l == r->objects[i]->id() && owner == r->objects[i]->owner())
		{
			int m = m_mirror[owner];
			int xx = x;
			int yy = y;
			if (m >=4)
			{
				xx = y;
				yy = x;
//				if (m == 4)
//					m = 0;
//				if (m == 5)
//					m = 2;
//				if (m == 6)
//					m = 1;
//				if (m == 7)
//					m = 3;
				m -= 4;
			}
			if (type.substr(0,7) == "ability" && type.length() == 8 && type[7] >= '0' && type[7] <= '9' && m_abilities_types[type[7]-'1'] == 'C')
			{
				(*(r->objects[i])).step(type,((m==0||m==2)?xx:m_maxx-xx-1), ((m==0||m==1)?yy:m_maxy-yy-1));
			}
			else
			{
				(*(r->objects[i])).step(type,((m==0||m==2)?xx:-xx),((m==0||m==1)?yy:-yy));
			}
			break;
		}
}


void	Server::score(int player, int n)
{
	m_round.back()->score[player]-=n;
	if (m_round.back()->score[player] - m_punishment_from_timeback[player] <= 0 && m_active_players[player])
	{
		m_active_players[player] = false;
		for (int i = 0; i < m_number_of_players; i++)
			m_next_player_info[i]->strings.push_back("A(z) "+itostr(player+1)+". játékos meghalt.");
//		m_round.back()->score[player] = 0;
	}
}
void	Server::leaver(int n)
{
       if (n < m_number_of_players)
       {
	       m_active_players[n] = false;
	       for (int i = 0; i < m_number_of_players; i++)
		       m_next_player_info[i]->strings.push_back("A(z) "+itostr(n+1)+". játékos kiszállt.");
       }
};
	

void	Server::check_game_over()
{
	int k = 0;
	int l = 0;
	for (int i = 0; i < m_number_of_players; i++)
		if (m_round.back()->score[i]-m_punishment_from_timeback[i] > 0 && m_active_players[i])
		{
			k++;
			l = i;
		}
	if (k == 1)
		for (int i = 0; i < m_number_of_players; i++)
			m_next_player_info[i]->strings.push_back("A(z) "+itostr(l+1)+". játékos nyert!!!");
	if (k == 0)
		for (int i = 0; i < m_number_of_players; i++)
			m_next_player_info[i]->strings.push_back("A játék döntetlennel zárult!");

	if (k <= 1)
		m_game_over = true;
}

int	Server::timeback(int number_of_rounds, int punishment, int for_who)
{
	m_punishment_from_timeback[for_who] += punishment;
	int r = m_copy_round_from - MAX(0,m_copy_round_from - number_of_rounds); 
	m_copy_round_from = MAX(0,m_copy_round_from - number_of_rounds);
	return r;
}
